#ifndef CAPD_H
#define CAPD_H

//some basic data structures, interval arithmetic, matrix and vector algorithms
#include "capd/interval/DoubleInterval.h"
#include "capd/vectalg/Vector.hpp"
#include "capd/vectalg/Matrix.hpp"
#include "capd/vectalg/Norm.hpp"

//Maps and functions representations
#include "capd/map/Function.hpp"
#include "capd/map/Map.hpp"
#include "capd/map/C2Map.hpp"

//ODE solvers
#include "capd/dynsys/BasicTaylor.hpp"
#include "capd/dynsys/C2Taylor.hpp"

//interval set representation, useful in ODE solvers, and checking cone conditions.
#include "capd/dynset/QRPolicy.h"
#include "capd/dynset/Intv2Set.hpp"
#include "capd/dynset/C1Rect.hpp"
#include "capd/dynset/C1Rect2.hpp"
#include "capd/dynset/Rect2Set.hpp"
#include "capd/dynset/RectSet.hpp"
#include "capd/dynset/Intv2Set.hpp"
#include "capd/dynset/C2Rect2.hpp"

//we will use poincare map to compute $M_{\pm}^{\pm}(\xi)$.
#include "capd/poincare/BasicPoincareMap.hpp"
#include "capd/poincare/PoincareMap.hpp"

//for exceptions in Sets
#include <sstream>
#include <string>
#include <stdexcept>


/** dimension of equation */
const int DIMENSION=2;
/** typedefs for short notation */
/** vector with double entries */
typedef capd::vectalg::Vector<double,DIMENSION> DoubleVector;  
/** vector with interval entries */
typedef capd::vectalg::Vector<interval,DIMENSION> IntervalVector;
/** vector with interval entries, one more entry for parameter */
typedef capd::vectalg::Vector<interval,DIMENSION+1> IntervalVectorParam;
/** matrix with double entries */
typedef capd::vectalg::Matrix<double,DIMENSION,DIMENSION> DoubleMatrix;
/** matrix with interval entries */
typedef capd::vectalg::Matrix<interval,DIMENSION,DIMENSION> IntervalMatrix;
/** matrix with interval entries, one more entry for parameter */
typedef capd::vectalg::Matrix<interval,DIMENSION+1,DIMENSION+1> IntervalMatrixParam;

typedef capd::vectalg::EuclNorm<IntervalVector, IntervalMatrix> EuclNorm;
typedef capd::vectalg::EuclLNorm<IntervalVector, IntervalMatrix> EuclLNorm;
typedef capd::vectalg::EuclNorm<IntervalVectorParam, IntervalMatrixParam> EuclNormParam;
typedef capd::vectalg::EuclLNorm<IntervalVectorParam, IntervalMatrixParam> EuclLNormParam;

/** Functions representation that can perform operations on intervals */
typedef capd::map::Function<IntervalVector> Function;
/** Functions representation that can perform operations on intervals, with one param */
typedef capd::map::Function<IntervalVectorParam> FunctionParam;
/** Map (i.e. function that goes from $\mathbb{R}^n \rightarrow \mathbb{R}^n$) representation that can perform operations on intervals */
typedef capd::map::Map<IntervalMatrix> Map;
/** Map (i.e. function that goes from $\mathbb{R}^n \rightarrow \mathbb{R}^n$) representation that can perform operations on intervals, with one param */
typedef capd::map::Map<IntervalMatrixParam> MapParam;

/** Dynamical system representation, i.e. differential equation */
typedef capd::dynsys::DynSys<IntervalMatrix> DynSys;
/** Dynamical system representation, i.e. differential equation, with one param */
typedef capd::dynsys::DynSys<IntervalMatrixParam> DynSysParam;
/** ODE solver */
typedef capd::dynsys::Taylor<Map> Taylor;
/** ODE solver, with one param */
typedef capd::dynsys::Taylor<MapParam> TaylorParam;

typedef capd::dynset::PartialQRWithPivoting<DIMENSION> QRPartial;

/** set types - used by Lohner-type algorithms*/
typedef capd::dynset::RectSet<IntervalMatrix> RectSet;
typedef capd::dynset::Rect2Set<IntervalMatrix> BaseRect2Set;

template<typename OutputType>
class StalkedRect2Set : public BaseRect2Set{
public:
	typedef BaseRect2Set BaseSet;
	typedef typename BaseSet::MatrixType MatrixType;
	typedef typename BaseSet::VectorType VectorType;
	typedef typename BaseSet::ScalarType ScalarType;
	typedef typename BaseSet::NormType NormType;
	typedef typename BaseSet::ReorganizationPolicy ReorganizationPolicy;
	typedef typename BaseSet::SetType SetType;
	typedef typename BaseSet::DynSysType DynSysType;	

	OutputType* out;
	interval forbiddenX;

	StalkedRect2Set(BaseSet & set) : BaseSet(set), out(0){}
	explicit StalkedRect2Set(int dim, double factor=1) : BaseSet(dim, factor), out(0) {}
	explicit StalkedRect2Set(const VectorType& x, double factor=1) : BaseSet(x, factor), out(0) {}
	StalkedRect2Set(const VectorType& x, const VectorType& r0, double factor=1) : BaseSet(x, r0, factor), out(0) {}
	StalkedRect2Set(const VectorType& x, const MatrixType& C, const VectorType& r0, double factor=1) : BaseSet(x, C, r0, factor), out(0) {}
	StalkedRect2Set(const VectorType& x, const MatrixType& C, const VectorType& r0, const VectorType& r, double factor=1) : BaseSet(x, C, r0, r, factor), out(0) {}
	StalkedRect2Set(const VectorType& x, const MatrixType& C, const VectorType& r0, const MatrixType& B,const VectorType& r, double factor=1) : BaseSet(x, C, r0, B, r, factor), out(0) {}

	std::string name() const { return "StalkedRect2Set"; }
	
	void setForbiddenX(interval argForbiddenX){
		this->forbiddenX = argForbiddenX;
	}

	void setOutput(OutputType& argout){
		out = &argout;
	}
	
	void move(DynSysType & dynsys){
		this->move(dynsys, *this);
	}

	void move(DynSysType & dynsys, StalkedRect2Set& result) const {
		//test if it do not cross line of singular points
		Taylor* method = dynamic_cast<Taylor*>(&dynsys);
		//current step of the method
		interval step = (method->getStep()).rightBound();
		//x0 - enclosure on the current set
		VectorType xx = m_x + m_C * m_r0 + m_B * m_r;
		//compute the rough enclosure for \varphi([0,step], xx)
		VectorType rough = capd::dynsys::enclosure(method->getField(), xx, step);
		//test if intersect with forbidden region 
		interval res;
		if (capd::intervals::intersection(forbiddenX, rough[0], res)){
		    std::ostringstream message;
			message << "Rect2Set.move(): possible nonempty intersection of the flow with the forbidden region: \nforbidden = " << forbiddenX << ", enclosure = " << rough << ", intersection = " << res;
			throw std::logic_error(message.str());
		}
		
		//move set
		BaseSet::move(dynsys, result);
		
		//output it (for visualizing)
		if (out){
			(*out) << result;
		}
	}
};

typedef capd::dynset::C1Rect2<IntervalMatrix> C1Rect2Set;

typedef capd::dynset::RectSet<IntervalMatrixParam, QRPartial> RectSetParam;
typedef capd::dynset::Rect2Set<IntervalMatrixParam, QRPartial> Rect2SetParam;

typedef capd::dynset::C1Rect2<IntervalMatrixParam, QRPartial> BaseC1Rect2SetParam;

template<typename OutputType>
class StalkedC1Rect2SetParam : public BaseC1Rect2SetParam{
public:
	typedef BaseC1Rect2SetParam BaseSet;
	typedef typename BaseSet::MatrixType MatrixType;
	typedef typename BaseSet::VectorType VectorType;
	typedef typename BaseSet::ScalarType ScalarType;
	typedef typename BaseSet::NormType NormType;
	typedef typename BaseSet::ReorganizationPolicy ReorganizationPolicy;
	typedef typename BaseSet::SetType SetType;
	typedef typename BaseSet::DynSysType DynSysType;	
		
	typedef typename BaseSet::SetType SetType;
	typedef typename BaseSet::C0BaseSet C0BaseSet;
	typedef typename BaseSet::C1BaseSet C1BaseSet;
	typedef bool C0Rect;
	typedef bool C1Rect;
	
	OutputType* out;	

	StalkedC1Rect2SetParam(const NormType& norm, int a) : BaseSet(norm, a), out(0){}
	StalkedC1Rect2SetParam(const VectorType& x, const NormType& n) : BaseSet(x, n), out(0){}
	StalkedC1Rect2SetParam(const VectorType& x, const VectorType& r, const NormType& n) : BaseSet(x, r, n), out(0){}
	StalkedC1Rect2SetParam(const VectorType& x, const MatrixType& B, const VectorType& r, const NormType& n) : BaseSet(x, B, r, n), out(0){}
	StalkedC1Rect2SetParam(const C1Rect2Set & set) : BaseSet(set), out(0){}

	StalkedC1Rect2SetParam &operator=(const VectorType &v){
		BaseSet::operator=(v);
		out = 0;
	}
	StalkedC1Rect2SetParam &operator=(const C1Rect2Set &S){
		BaseSet::operator=(S);
		out = 0;
		return *this;
	}
	StalkedC1Rect2SetParam &operator=(const StalkedC1Rect2SetParam &S){
		BaseSet::operator=(S);
		out = S.out;
		return *this;
	}

	std::string name() const { return "StalkedC1Rect2SetParam"; }

	void setOutput(OutputType& argout){
		out = &argout;
	}
	
	void move(DynSysType & dynsys){
		this->move(dynsys, *this);
	}

	void move(DynSysType & dynsys, StalkedC1Rect2SetParam& result) const {	  
		BaseSet::move(dynsys, result);
		if (out){
			(*out) << result;
		}
	}
};

/** For computing poincare map, we use it to simulate maps M_{\pm}^{\pm}(\xi) */
typedef capd::poincare::PoincareMap<Taylor> PoincareMap;
typedef capd::poincare::PoincareMap<TaylorParam> PoincareMapParam;
typedef capd::poincare::CrossingDirection CrossingDirection; 

#ifdef USE_GRAPHICS
#include "use_graphics.h"
typedef StalkedRect2Set<KrakOutput> Rect2Set;
typedef StalkedC1Rect2SetParam<KrakOutput> C1Rect2SetParam;
typedef KrakOutput GraphicsOutput;
#else
//we need dummy class
class Color{
public:
	Color(){}	
	static Color red;
	static Color green;
	static Color pine;
	static Color orange;
	static Color blue;
	static Color yellow;
};

Color Color::red;
Color Color::green;
Color Color::blue;
Color Color::orange;
Color Color::pine;
Color Color::yellow;

//we need dummy class
class Keypress{
public:
	std::string s;
	Keypress(std::string str) : s(str) {}
};

#include "output.h"
typedef StalkedRect2Set<NullOutput> Rect2Set;
typedef StalkedC1Rect2SetParam<NullOutput> C1Rect2SetParam;
typedef NullOutput GraphicsOutput;
#endif

#endif
